@@ -71,6 +71,8 @@
 #  define MIN(a,b) ((a)<(b)?(a):(b))
 #endif
 
+#define EOF_BLOCK "\037\213\010\4\0\0\0\0\0\377\6\0\102\103\2\0\033\0\3\0\0\0\0\0\0\0\0\0"
+
 /* Macros to store integers of various sizes in little endian byte order.
  * The value is put in the location pointed to by ucp, which should be 
  * an unsigned char pointer.  ucp is incremented by the size of the
@@ -195,7 +197,7 @@ static int bam_get_line(bam_file_t *b, unsigned char **str, size_t *len) {
 	 * On 64-bit OS this function becomes 3x faster.
 	 */
 #ifdef ALLOW_UAC
-#if SIZEOF_LONG == 8
+#if SIZEOF_LONG == 8 && ULONG_MAX != 0xffffffff
 #define hasless(x,n) (((x)-0x0101010101010101UL*(n))&~(x)&0x8080808080808080UL)
 #define haszero(x) (((x)-0x0101010101010101UL)&~(x)&0x8080808080808080UL)
 	{
@@ -258,7 +260,11 @@ static int bam_get_line(bam_file_t *b, unsigned char **str, size_t *len) {
 
     if (r == -1)
 	return -1;
-    return b->uncomp_sz ? -1 : 0;
+    if (b->uncomp_sz)
+	return -1;
+    
+    b->eof_block = 1; // expected eof
+    return 0;
 }
 
 static int load_bam_header(bam_file_t *b) {
@@ -523,8 +529,7 @@ int bam_close(bam_file_t *b) {
 	    BGZF_FLUSH(b);
 
 	    /* Output a blank BGZF block too to mark EOF */
-	    if (28 != fwrite("\037\213\010\4\0\0\0\0\0\377\6\0\102\103\2\0"
-			     "\033\0\3\0\0\0\0\0\0\0\0\0", 1, 28, b->fp)) {
+	    if (28 != fwrite(EOF_BLOCK, 1, 28, b->fp)) {
 		fprintf(stderr, "Write failed in bam_close()\n");
 	    }
 	} else {
@@ -680,10 +685,11 @@ static int bam_uncompress_input(bam_file_t *b) {
 		if (!(j = malloc(sizeof(*j))))
 		    return -1;
 
-		if (b->comp_sz < 18 && !b->eof) {
+	    empty_block_1:
+		if (b->comp_sz < 28 && !b->eof) {
 		    if (-1 == bam_more_input(b)) {
 			b->eof = 1;
-			if (b->comp_sz < 18) {
+			if (b->comp_sz < 28) {
 			    b->eof = 2;
 			    free(j);
 			    break;
@@ -695,6 +701,16 @@ static int bam_uncompress_input(bam_file_t *b) {
 		    break;
 		}
 	    
+		if (!b->eof) {
+		    if (memcmp(b->comp_p, EOF_BLOCK, 28) == 0) {
+			b->eof_block = 1;
+			b->comp_p += 28; b->comp_sz -= 28;
+			goto empty_block_1;
+		    } else {
+			b->eof_block = 0;
+		    }
+		}
+
 		bgzf = b->comp_p;
 		b->comp_p += 10; b->comp_sz -= 10;
 
@@ -800,12 +816,21 @@ static int bam_uncompress_input(bam_file_t *b) {
 
 	/* Uncompress another BGZF block */
 	/* BGZF header */
-	if (b->comp_sz < 18) {
+    empty_block_2:
+	if (b->comp_sz < 28) {
 	    if (-1 == bam_more_input(b))
 		return 0;
-	    if (b->comp_sz < 18)
+	    if (b->comp_sz < 28)
 		return -1;
 	}
+	
+	if (memcmp(b->comp_p, EOF_BLOCK, 28) == 0) {
+	    b->eof_block = 1;
+	    b->comp_p += 28; b->comp_sz -= 28;
+	    goto empty_block_2;
+	} else {
+	    b->eof_block = 0;
+	}
 
 	if (b->z_finish) {
 	    /*
@@ -929,7 +954,7 @@ static int bam_uncompress_input(bam_file_t *b) {
 }
 
 #ifdef ALLOW_UAC
-#if SIZEOF_LONG == 8
+#if SIZEOF_LONG == 8 && ULONG_MAX != 0xffffffff
 #define COPY_CPF_TO_CPTM(n)				\
     do {					        \
 	uint64_t *cpfi = (uint64_t *)cpf;		\
@@ -2925,6 +2950,11 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 
 	/* QNAME */
 	if (end - fp->uncomp_p < (sz = bam_name_len(b))) BF_FLUSH();
+	if (bam_name(b) - (char *)b + sz-1 >
+	    b->blk_size + offsetof(bam_seq_t, ref)) {
+	    fprintf(stderr, "Name length too large for bam block\n");
+	    return -1;
+	}
 	memcpy(fp->uncomp_p, bam_name(b), sz-1); fp->uncomp_p += sz-1;
 	*fp->uncomp_p++ = '\t';
 
@@ -3006,6 +3036,11 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 	n = (b->len+1)/2;
 	dat = (uc *)bam_seq(b);
 
+	if (dat - (uc *)b + b->len > b->blk_size + offsetof(bam_seq_t, ref)) {
+	    fprintf(stderr, "Sequence length too large for bam block\n");
+	    return -1;
+	}
+
 	/* BAM encoding */
 	//	while (n) {
 	//	    int l = end-fp->uncomp_p < n ? end-fp->uncomp_p : n;
@@ -3070,11 +3105,20 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 		dat += b->len;
 	    } else {
 		if (end - fp->uncomp_p < b->len + 3) BF_FLUSH();
-		if (end - fp->uncomp_p < b->len + 3) {
+		if (end - fp->uncomp_p < b->len + 3 ||
+		    fp->binning == BINNING_ILLUMINA) {
+
 		    /* Long seqs */
-		    for (i = 0; i < b->len; i++) {
-			if (end - fp->uncomp_p < 3) BF_FLUSH();
-			*fp->uncomp_p++ = *dat++ + '!';
+		    if (fp->binning == BINNING_ILLUMINA) {
+			for (i = 0; i < b->len; i++) {
+			    if (end - fp->uncomp_p < 3) BF_FLUSH();
+			    *fp->uncomp_p++ = illumina_bin_33[(uc)*dat++];
+			}
+		    } else {
+			for (i = 0; i < b->len; i++) {
+			    if (end - fp->uncomp_p < 3) BF_FLUSH();
+			    *fp->uncomp_p++ = *dat++ + '!';
+			}
 		    }
 		} else {
 		    unsigned char *cp = fp->uncomp_p;
@@ -3335,6 +3379,14 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
 	ptr        = (unsigned char *)cigar;
 #endif
 
+	if (fp->binning == BINNING_ILLUMINA) {
+	    int i;
+	    uc *q = (uc *)bam_qual(b);
+	    for (i = 0; i < b->len; i++) {
+		q[i] = illumina_bin[q[i]];
+	    }
+	}
+
         do {
             size_t blk_len = MIN(to_write, end - fp->uncomp_p);
             memcpy(fp->uncomp_p, ptr, blk_len);
@@ -3351,6 +3403,7 @@ int bam_put_seq(bam_file_t *fp, bam_seq_t *b) {
         } while(to_write > 0);
 
 #ifdef SP_BIG_ENDIAN
+	/* Swap back again */
 	b->ref         = le_int4(b->ref);
 	b->pos         = le_int4(b->pos);
 	b->bin_packed  = le_int4(b->bin_packed);
@@ -3479,6 +3532,10 @@ int bam_set_voption(bam_file_t *fd, enum bam_option opt, va_list args) {
 	fd->equeue = t_results_queue_init();
 	fd->dqueue = t_results_queue_init();
 	break;
+
+    case BAM_OPT_BINNING:
+	fd->binning = va_arg(args, int);
+	break;
     }
 
     return 0;
