@@ -53,6 +53,7 @@
 #include "io_lib/cram.h"
 #include "io_lib/os.h"
 #include "io_lib/md5.h"
+#include "io_lib/binning.h"
 
 #ifdef SAMTOOLS
 #    define bam_copy(dst, src) bam_copy1(*(dst), (src))
@@ -118,7 +119,7 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
      */
 
     // Duplicated from container itself, and removed in 1.1
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	itf8_put_blk(cb, h->ref_seq_id);
 	itf8_put_blk(cb, h->ref_seq_start);
 	itf8_put_blk(cb, h->ref_seq_span);
@@ -137,7 +138,7 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 	    return NULL;
 
 	hd.i = 1; HashTableAdd(h->preservation_map, "RN", 2, hd, NULL);
-	if (fd->version == CRAM_1_VERS) {
+	if (IS_CRAM_1_VERS(fd)) {
 	    hd.i = 0;
 	    if (!(HashTableAdd(h->preservation_map, "PI", 2, hd, NULL)))
 		return NULL;
@@ -391,7 +392,7 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 	    return NULL;
 	mc++;
     }
-    if (fd->version != CRAM_1_VERS) {
+    if (!IS_CRAM_1_VERS(fd)) {
 	if (h->SC_codec) {
 	    if (-1 == h->SC_codec->store(h->SC_codec, map, "SC", fd->version))
 		return NULL;
@@ -466,7 +467,7 @@ cram_block *cram_encode_compression_header(cram_fd *fd, cram_container *c,
 	    switch(hi->key[2]) {
 	    case 'Z': case 'H':
 		// string as byte_array_stop
-		if (fd->version == CRAM_1_VERS) {
+		if (IS_CRAM_1_VERS(fd)) {
 		    BLOCK_APPEND(map,
 				 "\005" // BYTE_ARRAY_STOP
 				 "\005" // len
@@ -602,7 +603,7 @@ cram_block *cram_encode_slice_header(cram_fd *fd, cram_slice *s) {
     cp += itf8_put(cp, s->hdr->ref_seq_start);
     cp += itf8_put(cp, s->hdr->ref_seq_span);
     cp += itf8_put(cp, s->hdr->num_records);
-    if (fd->version != CRAM_1_VERS)
+    if (!IS_CRAM_1_VERS(fd))
 	cp += itf8_put(cp, s->hdr->record_counter);
     cp += itf8_put(cp, s->hdr->num_blocks);
     cp += itf8_put(cp, s->hdr->num_content_ids);
@@ -612,7 +613,7 @@ cram_block *cram_encode_slice_header(cram_fd *fd, cram_slice *s) {
     if (s->hdr->content_type == MAPPED_SLICE)
 	cp += itf8_put(cp, s->hdr->ref_base_id);
 
-    if (fd->version != CRAM_1_VERS) {
+    if (!IS_CRAM_1_VERS(fd)) {
 	memcpy(cp, s->hdr->md5, 16); cp += 16;
     }
     
@@ -637,6 +638,8 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     int rec, r = 0, last_pos;
     cram_block *core;
     int nblk, embed_ref;
+    enum cram_block_method method1 = GZIP, method2 = RAW;
+    int level2, strat2 = Z_DEFAULT_STRATEGY;
 
     embed_ref = fd->embed_ref && s->hdr->ref_seq_id != -1 ? 1 : 0;
 
@@ -656,12 +659,12 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     s->hdr->ref_base_id = embed_ref ? CRAM_EXT_REF : -1;
     s->hdr->record_counter = c->num_records + c->record_counter;
     c->num_records += s->hdr->num_records;
-    nblk = (fd->version == CRAM_1_VERS) ? 5 : 6;
+    nblk = (IS_CRAM_1_VERS(fd)) ? 5 : 6;
 #ifdef BA_external
     nblk++;
 #endif
 #ifdef TN_external
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	nblk++;
     }
 #endif
@@ -681,12 +684,12 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     s->hdr->block_content_ids[3] = CRAM_EXT_TS_NP;
     s->hdr->block_content_ids[4] = CRAM_EXT_TAG;
     s->hdr->block_content_ids[5] = CRAM_EXT_SC;
-    nblk = (fd->version == CRAM_1_VERS) ? 5 : 6;
+    nblk = (IS_CRAM_1_VERS(fd)) ? 5 : 6;
 #ifdef BA_external
     s->hdr->block_content_ids[(s->ba_id = ++nblk)-1] = CRAM_EXT_BA;
 #endif
 #ifdef TN_external
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	s->hdr->block_content_ids[(s->tn_id = ++nblk)-1] = CRAM_EXT_TN;
     }
 #endif
@@ -699,7 +702,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     if (!(s->block[3] = cram_new_block(EXTERNAL, CRAM_EXT_NAME)))  return -1;
     if (!(s->block[4] = cram_new_block(EXTERNAL, CRAM_EXT_TS_NP))) return -1;
     if (!(s->block[5] = cram_new_block(EXTERNAL, CRAM_EXT_TAG)))   return -1;
-    if (fd->version != CRAM_1_VERS) {
+    if (!IS_CRAM_1_VERS(fd)) {
 	if (!(s->block[6] = cram_new_block(EXTERNAL, CRAM_EXT_SC)))
 	    return -1;
     }
@@ -708,7 +711,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 	return -1;
 #endif
 #ifdef TN_external
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	if (!(s->block[s->tn_id] = cram_new_block(EXTERNAL, CRAM_EXT_TN)))
 	    return -1;
     }
@@ -757,7 +760,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 	r |= h->CF_codec->encode(s, h->CF_codec, core,
 				 (char *)&i32, 1);
 
-	if (fd->version != CRAM_1_VERS)
+	if (!IS_CRAM_1_VERS(fd))
 	    r |= h->RI_codec->encode(s, h->RI_codec, core,
 				     (char *)&cr->ref_id, 1);
 
@@ -817,7 +820,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 	}
 
 	/* Aux tags */
-	if (fd->version == CRAM_1_VERS) {
+	if (IS_CRAM_1_VERS(fd)) {
 	    uc = cr->ntags;
 	    r |= h->TC_codec->encode(s, h->TC_codec, core, (char *)&uc, 1);
 #ifndef TN_external
@@ -968,7 +971,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     cram_free_block(s->block[2]);
     cram_free_block(s->block[3]);
     cram_free_block(s->block[5]);
-    if (fd->version != CRAM_1_VERS) {
+    if (!IS_CRAM_1_VERS(fd)) {
 	cram_free_block(s->block[6]);
 	BLOCK_UPLEN(s->soft_blk);
 	s->block[6] = s->soft_blk;
@@ -980,7 +983,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
     BLOCK_UPLEN(s->aux_blk);  s->block[5] = s->aux_blk;  s->aux_blk  = NULL;
 
 #ifdef TN_external
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	cram_free_block(s->block[s->tn_id]);
 	BLOCK_UPLEN(s->tn_blk); s->block[s->tn_id] = s->tn_blk;
 	s->tn_blk = NULL;
@@ -995,83 +998,110 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 
     /* Compress the CORE Block too, with minimal zlib level */
     if (fd->level > 5)
-	cram_compress_block(fd, s->block[0], NULL, 1, Z_CRAM_STRAT, -1, -1);
-
-#define USE_METRICS
-
-#ifdef USE_METRICS
-#  define LEVEL2 1
-#  define STRAT2 Z_RLE
-#else
-#  define LEVEL2 -1
-#  define STRAT2 -1
-#endif
+	cram_compress_block(fd, s->block[0], NULL,
+			    GZIP, 1, Z_CRAM_STRAT,
+			    RAW, 0, 0);
 
+    if (fd->use_bz2) {
+	method1 = BZIP2;
+	if (fd->use_arith) {
+	    method2 = RANS1;
+	    level2 = 1;
+	} else {
+	    method2 = GZIP;
+	    level2 = 3;
+	}
+    } else if (fd->use_arith) {
+	method1 = RANS1;
+	method2 = GZIP;
+	level2 = fd->level; // default
+	strat2 = Z_CRAM_STRAT;
+    } else {
+	method1 = GZIP;
+	method2 = GZIP;
+	level2 = 0; // disable except when level2 overridden
+	strat2 = Z_RLE;
+    }
     /* Compress the other blocks */
-    if (cram_compress_block(fd, s->block[1], NULL, //IN (seq)
-			    fd->level, Z_CRAM_STRAT,
-			    -1, -1))
+    if (cram_compress_block(fd, s->block[1], fd->m[0], //IN (seq)
+			    method1, fd->level, Z_CRAM_STRAT,
+			    method2, level2, strat2)) {
+	abort();
 	return -1;
+    }
 
     if (fd->level == 0) {
 	/* Do nothing */
     } else if (fd->level == 1) {
 	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
-				1, Z_RLE, -1, -1))
+				method1, 1, Z_RLE,
+				method2, level2,
+				method1 == GZIP ? strat2 : Z_RLE))
 	    return -1;
 	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
-				1, Z_RLE, -1, -1))
+				method1, 1, Z_RLE,
+				method2, level2, strat2))
 	    return -1;
     } else if (fd->level < 3) {
 	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
-				1, Z_RLE,
-				1, Z_HUFFMAN_ONLY))
+				method1, 1, Z_RLE,
+				method2, 1,
+				method1 == GZIP ? Z_HUFFMAN_ONLY : Z_RLE))
 	    return -1;
 	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
-				1, Z_RLE,
-				1, Z_HUFFMAN_ONLY))
+				method1, 1, Z_RLE,
+				method2, 1,
+				method1 == GZIP ? Z_HUFFMAN_ONLY : Z_RLE))
 	    return -1;
     } else {
 	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
-				fd->level, Z_CRAM_STRAT, 
-				LEVEL2, STRAT2))
+				method1, fd->level, Z_CRAM_STRAT, 
+				method2, -1, Z_RLE)) {
+	    abort();
 	    return -1;
+	}
 	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
-				fd->level, Z_CRAM_STRAT,
-				LEVEL2, STRAT2))
+				method1, fd->level, Z_CRAM_STRAT,
+				method2, level2, Z_DEFAULT_STRATEGY)) {
+	    abort();
 	    return -1;
+	}
     }
-    if (cram_compress_block(fd, s->block[3], NULL, //Name
-			    fd->level, Z_CRAM_STRAT,
-			    -1, -1))
-	return -1;
-    if (cram_compress_block(fd, s->block[4], NULL, //TS, NP
-			    fd->level, Z_CRAM_STRAT,
-			    -1, -1))
-	return -1;
-    if (fd->version != CRAM_1_VERS) {
-	if (cram_compress_block(fd, s->block[6], NULL, //SC (seq)
-				fd->level, Z_CRAM_STRAT,
-				-1, -1))
-	    return -1;
+
+    if (cram_compress_block(fd, s->block[3], fd->m[2], //Name
+			    method1, fd->level, Z_CRAM_STRAT,
+			    method2, level2, strat2))
+	abort();
+    if (cram_compress_block(fd, s->block[4], fd->m[3], //TS, NP
+			    method1, fd->level, Z_CRAM_STRAT,
+			    method2, level2, Z_DEFAULT_STRATEGY))
+	abort();
+    if (!IS_CRAM_1_VERS(fd)) {
+	if (cram_compress_block(fd, s->block[6], fd->m[5], //SC (seq)
+				method1, fd->level, Z_CRAM_STRAT,
+				method2, level2, strat2))
+	    abort();
     }
 #ifdef BA_external
-    if (cram_compress_block(fd, s->block[s->ba_id], NULL,
-			    fd->level, Z_CRAM_STRAT, -1, -1))
-	return -1;
+    if (cram_compress_block(fd, s->block[s->ba_id], fd->m[7],
+			    method1, fd->level, Z_CRAM_STRAT,
+			    method2, level2, strat2))
+	abort();
 #endif
 #ifdef TN_external
-    if (fd->version == CRAM_1_VERS) {
-	if (cram_compress_block(fd, s->block[s->tn_id], NULL,
-				fd->level, Z_DEFAULT_STRATEGY, -1, -1))
-	    return -1;
+    if (IS_CRAM_1_VERS(fd)) {
+	if (cram_compress_block(fd, s->block[s->tn_id], fd->m[8],
+				method1, fd->level, Z_DEFAULT_STRATEGY,
+				method2, level2, strat2))
+	    abort();
     }
 #endif
     if (embed_ref) {
 	BLOCK_UPLEN(s->block[s->ref_id]);
-	if (cram_compress_block(fd, s->block[s->ref_id], NULL,
-				fd->level, Z_DEFAULT_STRATEGY, -1, -1))
-	    return -1;
+	if (cram_compress_block(fd, s->block[s->ref_id], fd->m[9],
+				method1, fd->level, Z_DEFAULT_STRATEGY,
+				method2, level2, strat2))
+	    abort();
     }
 
     return r ? -1 : 0;
@@ -1208,7 +1238,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     for (i = 0; i < c->curr_slice; i++) {
 	cram_slice *s = c->slices[i];
 	
-	if (fd->version != CRAM_1_VERS) {
+	if (!IS_CRAM_1_VERS(fd)) {
 	    if (s->hdr->ref_seq_id >= 0 && c->multi_seq == 0 && !fd->no_ref) {
 		MD5_CTX md5;
 		MD5_Init(&md5);
@@ -1340,7 +1370,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 				    c->BS_stats, E_BYTE, NULL,
 				    fd->version);
 
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	h->TL_codec = NULL;
 	h->RI_codec = NULL;
 	h->RS_codec = NULL;
@@ -1464,7 +1494,7 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 	slice_offset = c_hdr->method == RAW
 	    ? c_hdr->uncomp_size
 	    : c_hdr->comp_size;
-	slice_offset += 2 +
+	slice_offset += 2 + 4*IS_CRAM_3_VERS(fd) +
 	    itf8_size(c_hdr->content_id) +
 	    itf8_size(c_hdr->comp_size) +
 	    itf8_size(c_hdr->uncomp_size);
@@ -1489,13 +1519,13 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
 	    ? s->hdr_block->uncomp_size
 	    : s->hdr_block->comp_size;
 
-	slice_offset += 2 + 
+	slice_offset += 2 + 4*IS_CRAM_3_VERS(fd) + 
 	    itf8_size(s->hdr_block->content_id) +
 	    itf8_size(s->hdr_block->comp_size) +
 	    itf8_size(s->hdr_block->uncomp_size);
 
 	for (j = 0; j < s->hdr->num_blocks; j++) {
-	    slice_offset += 2 + 
+	    slice_offset += 2 + 4*IS_CRAM_3_VERS(fd) + 
 		itf8_size(s->block[j]->content_id) +
 		itf8_size(s->block[j]->comp_size) +
 		itf8_size(s->block[j]->uncomp_size);
@@ -1569,6 +1599,9 @@ static int cram_add_substitution(cram_fd *fd, cram_container *c,
 	f.X.base = fd->cram_sub_matrix[ref&0x1f][base&0x1f];
 	cram_stats_add(c->BS_stats, f.X.base);
     } else {
+	if (fd->binning == BINNING_ILLUMINA)
+	    qual = illumina_bin[(uc)qual];
+
 	f.B.pos = pos+1;
 	f.B.code = 'B';
 	f.B.base = base;
@@ -1584,6 +1617,9 @@ static int cram_add_base(cram_fd *fd, cram_container *c,
 			 cram_slice *s, cram_record *r,
 			 int pos, char base, char qual) {
     cram_feature f;
+    if (fd->binning == BINNING_ILLUMINA)
+	qual = illumina_bin[(uc)qual];
+
     f.B.pos = pos+1;
     f.B.code = 'B';
     f.B.base = base;
@@ -1602,6 +1638,9 @@ static int cram_add_quality(cram_fd *fd, cram_container *c,
 			    cram_slice *s, cram_record *r,
 			    int pos, char qual) {
     cram_feature f;
+    if (fd->binning == BINNING_ILLUMINA)
+	qual = illumina_bin[(uc)qual];
+
     f.Q.pos = pos+1;
     f.Q.code = 'Q';
     f.Q.qual = qual;
@@ -1626,7 +1665,7 @@ static int cram_add_softclip(cram_container *c, cram_slice *s, cram_record *r,
     f.S.pos = pos+1;
     f.S.code = 'S';
     f.S.len = len;
-    if (version == CRAM_1_VERS) {
+    if (CRAM_MAJOR_VERS(version) == 1) {
 	f.S.seq_idx = BLOCK_SIZE(s->base_blk);
 	BLOCK_APPEND(s->base_blk, base, len);
 	BLOCK_APPEND_CHAR(s->base_blk, '\0');
@@ -2142,7 +2181,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     //cr->mate_flags;   // MF
     //cr->ntags;        // TC
     cr->ntags      = 0; //cram_stats_add(c->TC_stats, cr->ntags);
-    if (fd->version == CRAM_1_VERS)
+    if (IS_CRAM_1_VERS(fd))
 	rg = cram_encode_aux_1_0(fd, b, c, s, cr);
     else
 	rg = cram_encode_aux(fd, b, c, s, cr);
@@ -2155,7 +2194,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     if (rg) {
 	SAM_RG *brg = sam_hdr_find_rg(fd->header, rg);
 	cr->rg = brg ? brg->id : -1;
-    } else if (fd->version == CRAM_1_VERS) {
+    } else if (IS_CRAM_1_VERS(fd)) {
 	SAM_RG *brg = sam_hdr_find_rg(fd->header, "UNKNOWN");
 	assert(brg);
     } else {
@@ -2224,14 +2263,57 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     BLOCK_GROW(s->qual_blk, cr->len);
     seq = cp = (char *)BLOCK_END(s->seqs_blk);
 
-    *seq = 0;
-    for (i = 0; i < cr->len; i++) {
-	// FIXME: do 2 char at a time for efficiency
-#ifdef SAMTOOLS
-	cp[i] = seq_nt16_str[bam_seqi(bam_seq(b), i)];
-#else
-	cp[i] = bam_nt16_rev_table[bam_seqi(bam_seq(b), i)];
+    // Convert seq 2 bases at a time for speed.
+    {
+#ifdef ALLOW_UAC
+	static const uint16_t code2base[256] = {
+	    15677, 16701, 17213, 19773, 18237, 21053, 21309, 22077,
+	    21565, 22333, 22845, 18493, 19261, 17469, 16957, 20029,
+	    15681, 16705, 17217, 19777, 18241, 21057, 21313, 22081,
+	    21569, 22337, 22849, 18497, 19265, 17473, 16961, 20033,
+	    15683, 16707, 17219, 19779, 18243, 21059, 21315, 22083,
+	    21571, 22339, 22851, 18499, 19267, 17475, 16963, 20035,
+	    15693, 16717, 17229, 19789, 18253, 21069, 21325, 22093,
+	    21581, 22349, 22861, 18509, 19277, 17485, 16973, 20045,
+	    15687, 16711, 17223, 19783, 18247, 21063, 21319, 22087,
+	    21575, 22343, 22855, 18503, 19271, 17479, 16967, 20039,
+	    15698, 16722, 17234, 19794, 18258, 21074, 21330, 22098,
+	    21586, 22354, 22866, 18514, 19282, 17490, 16978, 20050,
+	    15699, 16723, 17235, 19795, 18259, 21075, 21331, 22099,
+	    21587, 22355, 22867, 18515, 19283, 17491, 16979, 20051,
+	    15702, 16726, 17238, 19798, 18262, 21078, 21334, 22102,
+	    21590, 22358, 22870, 18518, 19286, 17494, 16982, 20054,
+	    15700, 16724, 17236, 19796, 18260, 21076, 21332, 22100,
+	    21588, 22356, 22868, 18516, 19284, 17492, 16980, 20052,
+	    15703, 16727, 17239, 19799, 18263, 21079, 21335, 22103,
+	    21591, 22359, 22871, 18519, 19287, 17495, 16983, 20055,
+	    15705, 16729, 17241, 19801, 18265, 21081, 21337, 22105,
+	    21593, 22361, 22873, 18521, 19289, 17497, 16985, 20057,
+	    15688, 16712, 17224, 19784, 18248, 21064, 21320, 22088,
+	    21576, 22344, 22856, 18504, 19272, 17480, 16968, 20040,
+	    15691, 16715, 17227, 19787, 18251, 21067, 21323, 22091,
+	    21579, 22347, 22859, 18507, 19275, 17483, 16971, 20043,
+	    15684, 16708, 17220, 19780, 18244, 21060, 21316, 22084,
+	    21572, 22340, 22852, 18500, 19268, 17476, 16964, 20036,
+	    15682, 16706, 17218, 19778, 18242, 21058, 21314, 22082,
+	    21570, 22338, 22850, 18498, 19266, 17474, 16962, 20034,
+	    15694, 16718, 17230, 19790, 18254, 21070, 21326, 22094,
+	    21582, 22350, 22862, 18510, 19278, 17486, 16974, 20046
+	};
 #endif
+	int l2 = cr->len & ~1;
+	unsigned char *from = bam_seq(b);
+	cp[0] = 0;
+	for (i = 0; i < l2; i += 2, from++) {
+#ifdef ALLOW_UAC
+	    *(int16_t *)&cp[i] = le_int2(code2base[*from]);
+#else
+	    cp[i+0] = bam_nt16_rev_table[*from >> 4];
+	    cp[i+1] = bam_nt16_rev_table[*from & 0xf];
+#endif
+	}
+	if (i < cr->len)
+	    cp[i] = bam_nt16_rev_table[*from >> 4];
     }
     BLOCK_SIZE(s->seqs_blk) += cr->len;
 
@@ -2276,16 +2358,20 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 		if (!fd->no_ref && cr->len) {
 		    int end = cig_len+apos < c->ref_end
 			? cig_len : c->ref_end - apos;
-		    for (l = 0; l < end && seq[spos]; l++, apos++, spos++) {
-			if (ref[apos] != seq[spos]) {
-			    //fprintf(stderr, "Subst: %d; %c vs %c\n",
-			    //	spos, ref[apos], seq[spos]);
-			    if (cram_add_substitution(fd, c, s, cr, spos,
-						      seq[spos], qual[spos],
-						      ref[apos]))
+		    char *sp = &seq[spos];
+		    char *rp = &ref[apos];
+		    char *qp = &qual[spos];
+		    for (l = 0; l < end; l++) {
+			if (rp[l] != sp[l]) {
+			    if (!sp[l])
+				break;
+			    if (cram_add_substitution(fd, c, s, cr, spos+l,
+						      sp[l], qp[l], rp[l]))
 				return -1;
 			}
 		    }
+		    spos += l;
+		    apos += l;
 		}
 
 		if (l < cig_len && cr->len) {
@@ -2394,7 +2480,14 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	    cp = (char *)BLOCK_END(s->qual_blk);
 	    char *from = &bam_qual(b)[0];
 	    char *to = &cp[0];
-	    memcpy(to, from, cr->len);
+
+	    if (fd->binning == BINNING_ILLUMINA) {
+		int i;
+		for (i = 0; i < cr->len; i++)
+		    to[i] = illumina_bin[(uc)from[i]];
+	    } else {
+		memcpy(to, from, cr->len);
+	    }
 	    //for (i = 0; i < cr->len; i++) cp[i] = from[i];
 	}
 	BLOCK_SIZE(s->qual_blk) += cr->len;
@@ -2553,7 +2646,7 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	slice_rec = c->slice_rec;
 	curr_rec  = c->curr_rec;
 
-	if (fd->version == CRAM_1_VERS ||
+	if (IS_CRAM_1_VERS(fd) ||
 	    c->curr_rec == c->max_rec || fd->multi_seq != 1 || !c->slice)
 	    if (NULL == (c = cram_next_container(fd, b)))
 		return -1;
