@@ -62,6 +62,9 @@
 #ifdef HAVE_LIBBZ2
 #include <bzlib.h>
 #endif
+#ifdef HAVE_LIBLZMA
+#include <lzma.h>
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <math.h>
@@ -71,17 +74,19 @@
 #include "io_lib/os.h"
 #include "io_lib/md5.h"
 #include "io_lib/open_trace_file.h"
+#include "io_lib/arith_static.h"
+#include "io_lib/rANS_static.h"
 
 //#define REF_DEBUG
 
 #ifdef REF_DEBUG
 #define RP(...) fprintf (stderr, __VA_ARGS__)
-#else
-#define RP(...) 
-#endif
-
 #include <sys/syscall.h>
 #define gettid() (int)syscall(SYS_gettid)
+#else
+#define RP(...) 
+#define gettid() 0
+#endif
 
 
 /* ----------------------------------------------------------------------
@@ -519,15 +524,14 @@ int int32_put(cram_block *b, int32_t val) {
  * They're static here as they're only used within the cram_compress_block
  * and cram_uncompress_block functions, which are the external interface.
  */
-
 static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
     z_stream s;
     unsigned char *data = NULL; /* Uncompressed output */
     int data_alloc = 0;
     int err;
 
-    /* Starting point at uncompressed size, 4x compressed */
-    data = malloc(data_alloc = csize+10);
+    /* Starting point at uncompressed size, and scale after that */
+    data = malloc(data_alloc = csize*1.2+100);
     if (!data)
 	return NULL;
 
@@ -553,6 +557,7 @@ static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
     /* Decode to 'data' array */
     for (;s.avail_in;) {
 	unsigned char *data_tmp;
+	int alloc_inc;
 
 	s.next_out = &data[s.total_out];
 	err = inflate(&s, Z_NO_FLUSH);
@@ -564,13 +569,14 @@ static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
 	    break;
 	}
 
-	/* More to come, so realloc */
-	data = realloc((data_tmp = data), data_alloc += s.avail_in + 10);
+	/* More to come, so realloc based on growth so far */
+	alloc_inc = (double)s.avail_in/s.total_in * s.total_out + 100;
+	data = realloc((data_tmp = data), data_alloc += alloc_inc);
 	if (!data) {
 	    free(data_tmp);
 	    return NULL;
 	}
-	s.avail_out += s.avail_in+10;
+	s.avail_out += alloc_inc;
     }
     inflateEnd(&s);
 
@@ -635,6 +641,90 @@ static char *zlib_mem_deflate(char *data, size_t size, size_t *cdata_size,
     return (char *)cdata;
 }
 
+#ifdef HAVE_LIBLZMA
+/* ------------------------------------------------------------------------ */
+/*
+ * Data compression routines using liblzma (xz)
+ *
+ * On a test set this shrunk the main db from 136157104 bytes to 114796168, but
+ * caused tg_index to grow from 2m43.707s to 15m3.961s. Exporting as bfastq
+ * went from 18.3s to 36.3s. So decompression suffers too, but not as bad
+ * as compression times.
+ *
+ * For now we disable this functionality. If it's to be reenabled make sure you
+ * improve the mem_inflate implementation as it's just a test hack at the
+ * moment.
+ */
+
+static char *lzma_mem_deflate(char *data, size_t size, size_t *cdata_size,
+			      int level) {
+    char *out;
+    size_t out_size = lzma_stream_buffer_bound(size);
+    *cdata_size = 0;
+
+    out = malloc(out_size);
+
+    /* Single call compression */
+    if (LZMA_OK != lzma_easy_buffer_encode(level, LZMA_CHECK_CRC32, NULL,
+					   (uint8_t *)data, size,
+					   (uint8_t *)out, cdata_size,
+					   out_size))
+    	return NULL;
+
+    return out;
+}
+
+static char *lzma_mem_inflate(char *cdata, size_t csize, size_t *size) {
+    lzma_stream strm = LZMA_STREAM_INIT;
+    size_t out_size = 0, out_pos = 0;
+    char *out = NULL;
+    int r;
+
+    /* Initiate the decoder */
+    if (LZMA_OK != lzma_stream_decoder(&strm, 50000000, 0))
+	return NULL;
+
+    /* Decode loop */
+    strm.avail_in = csize;
+    strm.next_in = (uint8_t *)cdata;
+
+    for (;strm.avail_in;) {
+	if (strm.avail_in > out_size - out_pos) {
+	    out_size += strm.avail_in * 4 + 32768;
+	    out = realloc(out, out_size);
+	}
+	strm.avail_out = out_size - out_pos;
+	strm.next_out = (uint8_t *)&out[out_pos];
+
+	r = lzma_code(&strm, LZMA_RUN);
+	if (LZMA_OK != r && LZMA_STREAM_END != r) {
+	    fprintf(stderr, "r=%d\n", r);
+	    fprintf(stderr, "mem=%"PRId64"d\n", (int64_t)lzma_memusage(&strm));
+	    return NULL;
+	}
+
+	out_pos = strm.total_out;
+
+	if (r == LZMA_STREAM_END)
+	    break;
+    }
+
+    /* finish up any unflushed data; necessary? */
+    r = lzma_code(&strm, LZMA_FINISH);
+    if (r != LZMA_OK && r != LZMA_STREAM_END) {
+	fprintf(stderr, "r=%d\n", r);
+	return NULL;
+    }
+
+    out = realloc(out, strm.total_out);
+    *size = strm.total_out;
+
+    lzma_end(&strm);
+
+    return out;
+}
+#endif
+
 /* ----------------------------------------------------------------------
  * CRAM blocks - the dynamically growable data block. We have code to
  * create, update, (un)compress and read/write.
@@ -707,6 +797,29 @@ cram_block *cram_read_block(cram_fd *fd) {
 	}
     }
 
+    if (IS_CRAM_3_VERS(fd)) {
+	unsigned char dat[100], *cp = dat;;
+	uint32_t crc;
+
+	
+	if (-1 == int32_decode(fd, &b->crc32))  { free(b); return NULL; }
+
+	*cp++ = b->method;
+	*cp++ = b->content_type;
+	cp += itf8_put(cp, b->content_id);
+	cp += itf8_put(cp, b->comp_size);
+	cp += itf8_put(cp, b->uncomp_size);
+	crc = crc32(0L, dat, cp-dat);
+	crc = crc32(crc, b->data ? b->data : (uc *)"", b->alloc);
+
+	if (crc != b->crc32) {
+	    fprintf(stderr, "Block CRC32 failure\n");
+	    free(b->data);
+	    free(b);
+	    return NULL;
+	}
+    }
+
     b->orig_method = b->method;
     b->idx = 0;
     b->byte = 0;
@@ -737,6 +850,27 @@ int cram_write_block(cram_fd *fd, cram_block *b) {
 	    return -1;
     }
 
+    if (IS_CRAM_3_VERS(fd)) {
+	unsigned char dat[100], *cp = dat;;
+	uint32_t crc;
+
+	*cp++ = b->method;
+	*cp++ = b->content_type;
+	cp += itf8_put(cp, b->content_id);
+	cp += itf8_put(cp, b->comp_size);
+	cp += itf8_put(cp, b->uncomp_size);
+	crc = crc32(0L, dat, cp-dat);
+
+	if (b->method == RAW) {
+	    b->crc32 = crc32(crc, b->data ? b->data : (uc*)"", b->uncomp_size);
+	} else {
+	    b->crc32 = crc32(crc, b->data ? b->data : (uc*)"", b->comp_size);
+	}
+
+	if (-1 == int32_encode(fd, b->crc32))
+	    return -1;
+    }
+
     return 0;
 }
 
@@ -777,6 +911,7 @@ int cram_uncompress_block(cram_block *b) {
 	    return -1;
 	free(b->data);
 	b->data = (unsigned char *)uncomp;
+	b->alloc = uncomp_size;
 	b->method = RAW;
 	break;
 
@@ -792,6 +927,7 @@ int cram_uncompress_block(cram_block *b) {
 	    return -1;
 	}
 	b->data = (unsigned char *)uncomp;
+	b->alloc = usize;
 	b->method = RAW;
 	b->uncomp_size = usize; // Just incase it differs
 	break;
@@ -804,6 +940,74 @@ int cram_uncompress_block(cram_block *b) {
 	break;
 #endif
 
+#ifdef HAVE_LIBLZMA
+    case LZMA:
+	uncomp = lzma_mem_inflate((char *)b->data, b->comp_size, &uncomp_size);
+	if (!uncomp)
+	    return -1;
+	if ((int)uncomp_size != b->uncomp_size)
+	    return -1;
+	free(b->data);
+	b->data = (unsigned char *)uncomp;
+	b->alloc = uncomp_size;
+	b->method = RAW;
+	break;
+#else
+    case LZMA:
+	fprintf(stderr, "Lzma compression is not compiled into this "
+		"version.\nPlease rebuild and try again.\n");
+	return -1;
+	break;
+#endif
+
+    case ARITH0: {
+	unsigned int usize = b->uncomp_size, usize2;
+	uncomp = arith_uncompress(b->data, b->comp_size, &usize2, 0);
+	assert(usize == usize2);
+	b->data = (unsigned char *)uncomp;
+	b->alloc = usize2;
+	b->method = RAW;
+	b->uncomp_size = usize2; // Just incase it differs
+	//fprintf(stderr, "Expanded %d to %d\n", b->comp_size, b->uncomp_size);
+	break;
+    }
+
+    case ARITH1: {
+	unsigned int usize = b->uncomp_size, usize2;
+	uncomp = arith_uncompress(b->data, b->comp_size, &usize2, 1);
+	assert(usize == usize2);
+	b->data = (unsigned char *)uncomp;
+	b->alloc = usize2;
+	b->method = RAW;
+	b->uncomp_size = usize2; // Just incase it differs
+	//fprintf(stderr, "Expanded %d to %d\n", b->comp_size, b->uncomp_size);
+	break;
+    }
+
+    case RANS0: {
+	unsigned int usize = b->uncomp_size, usize2;
+	uncomp = rans_uncompress(b->data, b->comp_size, &usize2, 0);
+	assert(usize == usize2);
+	b->data = (unsigned char *)uncomp;
+	b->alloc = usize2;
+	b->method = RAW;
+	b->uncomp_size = usize2; // Just incase it differs
+	//fprintf(stderr, "Expanded %d to %d\n", b->comp_size, b->uncomp_size);
+	break;
+    }
+
+    case RANS1: {
+	unsigned int usize = b->uncomp_size, usize2;
+	uncomp = rans_uncompress(b->data, b->comp_size, &usize2, 1);
+	assert(usize == usize2);
+	b->data = (unsigned char *)uncomp;
+	b->alloc = usize2;
+	b->method = RAW;
+	b->uncomp_size = usize2; // Just incase it differs
+	//fprintf(stderr, "Expanded %d to %d\n", b->comp_size, b->uncomp_size);
+	break;
+    }
+
     default:
 	return -1;
     }
@@ -811,38 +1015,135 @@ int cram_uncompress_block(cram_block *b) {
     return 0;
 }
 
-#ifdef HAVE_LIBBZ2
-static int cram_compress_block_bzip2(cram_fd *fd, cram_block *b,
-				     cram_metrics *metrics, int level) {
-    unsigned int comp_size = b->uncomp_size*1.01 + 600;
-    char *comp = malloc(comp_size);
-    char *data = (char *)b->data;
+#define EBASE 65536
+static double entropy16(unsigned short *data, int len) {
+    double E[EBASE];
+    double P[EBASE];
+    double e = 0;
+    int i;
+    
+    for (i = 0; i < EBASE; i++)
+        P[i] = 0;
 
-    if (!comp)
-	return -1;
+    for (i = 0; i < len; i++)
+        P[data[i]]++;
 
-    if (!data)
-	data = "";
-
-    if (BZ_OK != BZ2_bzBuffToBuffCompress(comp, &comp_size,
-					  data, b->uncomp_size,
-					  level, 0, 30)) {
-	free(comp);
-	return -1;
+    for (i = 0; i < EBASE; i++) {
+        if (P[i]) {
+            P[i] /= len;
+            E[i] = -(log(P[i])/log(EBASE));
+        } else {
+            E[i] = 0;
+        }
     }
 
-    free(b->data);
-    b->data = (unsigned char *)comp;
-    b->method = BZIP2;
-    b->comp_size = comp_size;
+    for (e = i = 0; i < len; i++)
+        e += E[data[i]];
 
-    if (fd->verbose)
-	fprintf(stderr, "Compressed block ID %d from %d to %d\n",
-		b->content_id, b->uncomp_size, b->comp_size);
-
-    return 0;
+    return e * log(EBASE)/log(256);
 }
+
+#define EBASE2 256
+static double entropy8(unsigned char *data, int len) {
+    int F[EBASE2];
+    double e = 0;
+    int i;
+    
+    for (i = 0; i < EBASE2; i++)
+        F[i] = 0;
+
+    for (i = 0; i < len; i++)
+        F[data[i]]++;
+
+    for (i = 0; i < EBASE2; i++) {
+        if (F[i]) {
+	    e += -log((double)F[i]/len) * F[i];
+        }
+    }
+
+    return e / log(EBASE2);
+}
+
+static char *cram_compress_by_method(char *in, size_t in_size,
+				     size_t *out_size,
+				     enum cram_block_method method,
+				     int level, int strat) {
+    switch (method) {
+    case GZIP:
+	return zlib_mem_deflate(in, in_size, out_size, level, strat);
+
+    case BZIP2: {
+#ifdef HAVE_LIBBZ2
+	unsigned int comp_size = in_size*1.01 + 600;
+	char *comp = malloc(comp_size);
+	if (!comp)
+	    return NULL;
+
+	if (BZ_OK != BZ2_bzBuffToBuffCompress(comp, &comp_size,
+					      in, in_size,
+					      level, 0, 30)) {
+	    free(comp);
+	    return NULL;
+	}
+	*out_size = comp_size;
+	return comp;
+#else
+	return NULL;
 #endif
+    }
+
+    case LZMA:
+#ifdef HAVE_LIBLZMA
+	return lzma_mem_deflate(in, in_size, out_size, level);
+#else
+	return NULL;
+#endif
+
+    case ARITH0: {
+	unsigned int out_size_i;
+	unsigned char *cp;
+	cp = arith_compress(in, in_size, &out_size_i, 0);
+	*out_size = out_size_i;
+	return cp;
+    }
+
+    case ARITH1: {
+	unsigned int out_size_i;
+	unsigned char *cp;
+	
+//	double e8  = entropy8(in, in_size);
+//	double e16 = entropy16((unsigned short *)in, in_size/2);
+//
+//	if (e8 / e16 < 1.02 || e8 - e16 < 512) {
+//	    return cram_compress_by_method(in, in_size, out_size,
+//					   ARITH0, level, strat);
+//	}
+
+	cp = arith_compress(in, in_size, &out_size_i, 1);
+	*out_size = out_size_i;
+	return cp;
+    }
+
+    case RANS0: {
+	unsigned int out_size_i;
+	unsigned char *cp;
+	cp = rans_compress(in, in_size, &out_size_i, 0);
+	*out_size = out_size_i;
+	return cp;
+    }
+
+    case RANS1: {
+	unsigned int out_size_i;
+	unsigned char *cp;
+	
+	cp = rans_compress(in, in_size, &out_size_i, 1);
+	*out_size = out_size_i;
+	return cp;
+    }
+    }
+
+    return NULL;
+}
 
 /*
  * Compresses a block using one of two different zlib strategies. If we only
@@ -853,12 +1154,14 @@ static int cram_compress_block_bzip2(cram_fd *fd, cram_block *b,
  * significantly faster.
  */
 int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
-			int level,  int strat,
-			 int level2, int strat2) {
+			enum cram_block_method method1, int level1, int strat1,
+			enum cram_block_method method2, int level2,int strat2){
+
     char *comp = NULL;
     size_t comp_size = 0;
+    enum cram_block_method method;
 
-    if (level == 0) {
+    if (method1 == RAW || level1 == 0 || b->uncomp_size == 0) {
 	b->method = RAW;
 	b->comp_size = b->uncomp_size;
 	return 0;
@@ -869,22 +1172,14 @@ int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
 	return 0;
     }
 
-#ifdef HAVE_LIBBZ2
-    if (fd->use_bz2)
-	// metrics ignored for bzip2
-	return cram_compress_block_bzip2(fd, b, metrics, level);
-#endif
-    
-    //return cram_compress_block_arith1(fd, b, metrics);
-
     pthread_mutex_lock(&fd->metrics_lock);
-    if (strat2 >= 0)
+    if (level2)
 	if (fd->verbose > 1)
 	    fprintf(stderr, "metrics trial %d, next_trial %d, m1 %d, m2 %d\n",
 		    metrics->trial, metrics->next_trial,
 		    metrics->m1, metrics->m2);
-
-    if (strat2 >= 0 && (metrics->trial || --metrics->next_trial == 0)) {
+    
+    if (level2 && (metrics->trial > 0 || --metrics->next_trial <= 0)) {
 	char *c1, *c2;
 	size_t s1, s2;
 
@@ -895,10 +1190,10 @@ int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
 	}
 	pthread_mutex_unlock(&fd->metrics_lock);
 
-	c1 = zlib_mem_deflate((char *)b->data, b->uncomp_size,
-			      &s1, level, strat);
-	c2 = zlib_mem_deflate((char *)b->data, b->uncomp_size,
-			      &s2, level2, strat2);
+	c1 = cram_compress_by_method((char *)b->data, b->uncomp_size, &s1,
+				     method1, level1, strat1);
+	c2 = cram_compress_by_method((char *)b->data, b->uncomp_size, &s2,
+				     method2, level2, strat2);
 	if (!c1 || !c2)
 	    return -1;
 	
@@ -911,25 +1206,30 @@ int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
 	    comp = c1; comp_size = s1;
 	    free(c2);
 	    metrics->m1++;
+	    method = method1;
 	} else {
 	    if (fd->verbose > 1)
 		fprintf(stderr, "M2 wins %d vs %d\n", (int)s1, (int)s2);
 	    comp = c2; comp_size = s2;
 	    free(c1);
 	    metrics->m2++;
+	    method = method2;
 	}
 	metrics->trial--;
 	pthread_mutex_unlock(&fd->metrics_lock);
-    } else if (strat2 >= 0) {
-	int xlevel = metrics->m1 > metrics->m2 ? level : level2;
-	int xstrat = metrics->m1 > metrics->m2 ? strat : strat2;
+    } else if (level2) {
+	int xlevel = metrics->m1 > metrics->m2 ? level1 : level2;
+	int xstrat = metrics->m1 > metrics->m2 ? strat1 : strat2;
+
+	method = metrics->m1 > metrics->m2 ? method1 : method2;
 	pthread_mutex_unlock(&fd->metrics_lock);
-	comp = zlib_mem_deflate((char *)b->data, b->uncomp_size, &comp_size,
-				xlevel, xstrat);
+	comp = cram_compress_by_method((char *)b->data, b->uncomp_size,
+				       &comp_size, method, xlevel, xstrat);
     } else {
 	pthread_mutex_unlock(&fd->metrics_lock);
-	comp = zlib_mem_deflate((char *)b->data, b->uncomp_size, &comp_size,
-				level, strat);
+	comp = cram_compress_by_method((char *)b->data, b->uncomp_size,
+				       &comp_size, method1, level1, strat1);
+	method = method1;
     }
 
     if (!comp)
@@ -937,12 +1237,13 @@ int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
 
     free(b->data);
     b->data = (unsigned char *)comp;
-    b->method = GZIP;
+    b->method = method;
     b->comp_size = comp_size;
 
     if (fd->verbose)
-	fprintf(stderr, "Compressed block ID %d from %d to %d\n",
-		b->content_id, b->uncomp_size, b->comp_size);
+	fprintf(stderr, "Compressed block ID %d from %d to %d by method %s\n",
+		b->content_id, b->uncomp_size, b->comp_size,
+		cram_block_method2str(b->method));
 
     return 0;
 }
@@ -962,6 +1263,11 @@ char *cram_block_method2str(enum cram_block_method m) {
     case RAW:	return "RAW";
     case GZIP:	return "GZIP";
     case BZIP2:	return "BZIP2";
+    case LZMA:  return "LZMA";
+    case ARITH0:return "ARITH0";
+    case ARITH1:return "ARITH1";
+    case RANS0: return "RANS0";
+    case RANS1: return "RANS1";
     }
     return "?";
 }
@@ -992,12 +1298,15 @@ int paranoid_fclose(FILE *fp) {
     }
 
     errno = 0;
+#ifdef HAVE_FSYNC
     if (-1 == fsync(fileno(fp))) {
 	if (errno != EINVAL) { // eg pipe
 	    fclose(fp);
 	    return -1;
 	}
     }
+#endif
+
     return fclose(fp);
 }
 
@@ -1078,6 +1387,7 @@ static refs_t *refs_create(void) {
     r->ref_id = NULL; // see refs2id() to populate.
     r->count = 1;
     r->last = NULL;
+    r->last_id = -1;
 
     r->h_meta = HashTableCreate(16, HASH_DYNAMIC_SIZE | HASH_NONVOLATILE_KEYS);
     if (!r->h_meta)
@@ -1451,13 +1761,9 @@ static int cram_populate_ref(cram_fd *fd, int id, ref_entry *r) {
 
     /* Otherwise search */
     if ((mf = open_path_mfile(tag->str+3, ref_path, NULL))) {
-	mfseek(mf, 0, SEEK_END);
-	r->length = mftell(mf);
-	r->seq = malloc(r->length);
-	mrewind(mf);
-	mfread(r->seq, 1, r->length, mf);
-	mfclose(mf);
-
+	size_t sz;
+	r->seq = mfsteal(mf, &sz);
+	r->length = sz;
     } else {
 	refs_t *refs;
 	char *fn;
@@ -1540,6 +1846,9 @@ static void cram_ref_incr_locked(refs_t *r, int id) {
     if (id < 0 || !r->ref_id[id]->seq)
 	return;
 
+    if (r->last_id == id)
+	r->last_id = -1;
+
     ++r->ref_id[id]->count;
 }
 
@@ -1559,11 +1868,18 @@ static void cram_ref_decr_locked(refs_t *r, int id) {
 
     if (--r->ref_id[id]->count <= 0) {
 	assert(r->ref_id[id]->count == 0);
-	RP("%d FREE REF %d (%p)\n", gettid(), id, r->ref_id[id]->seq);
-	if (r->ref_id[id]->seq) {
-	    free(r->ref_id[id]->seq);
-	    r->ref_id[id]->seq = NULL;
-	    r->ref_id[id]->length = 0;
+	if (r->last_id >= 0) {
+	    if (r->ref_id[r->last_id]->count <= 0 &&
+		r->ref_id[r->last_id]->seq) {
+		RP("%d FREE REF %d (%p)\n", gettid(),
+		   r->last_id, r->ref_id[r->last_id]->seq);
+		free(r->ref_id[r->last_id]->seq);
+		r->ref_id[r->last_id]->seq = NULL;
+		r->ref_id[r->last_id]->length = 0;
+	    }
+	    r->last_id = -1;
+	} else {
+	    r->last_id = id;
 	}
     }
 }
@@ -1636,6 +1952,11 @@ static char *load_ref_portion(FILE *fp, ref_entry *e, int start, int end) {
 	    free(seq);
 	    return NULL;
 	}
+    } else {
+	int i;
+	for (i = 0; i < len; i++) {
+	    seq[i] = seq[i] & ~0x20; // uppercase in ASCII
+	}
     }
 
     return seq;
@@ -2112,16 +2433,16 @@ cram_container *cram_read_container(cram_fd *fd) {
     fd->err = 0;
 
     memset(&c2, 0, sizeof(c2));
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	if ((s = itf8_decode(fd, &c2.length)) == -1) {
-	    fd->eof = 1;
+	    fd->eof = fd->empty_container ? 1 : 2;
 	    return NULL;
 	} else {
 	    rd+=s;
 	}
     } else {
 	if ((s = int32_decode(fd, &c2.length)) == -1) {
-	    fd->eof = 1;
+	    fd->eof = fd->empty_container ? 1 : 2;
 	    return NULL;
 	} else {
 	    rd+=s;
@@ -2132,7 +2453,7 @@ cram_container *cram_read_container(cram_fd *fd) {
     if ((s = itf8_decode(fd, &c2.ref_seq_span)) == -1) return NULL; else rd+=s;
     if ((s = itf8_decode(fd, &c2.num_records))  == -1) return NULL; else rd+=s;
 
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	c2.record_counter = 0;
 	c2.num_bases = 0;
     } else {
@@ -2154,7 +2475,8 @@ cram_container *cram_read_container(cram_fd *fd) {
 
     *c = c2;
 
-    if (!(c->landmark = malloc(c->num_landmarks * sizeof(int32_t)))) {
+    if (!(c->landmark = malloc(c->num_landmarks * sizeof(int32_t))) &&
+	c->num_landmarks) {
 	fd->err = errno;
 	cram_free_container(c);
 	return NULL;
@@ -2167,8 +2489,49 @@ cram_container *cram_read_container(cram_fd *fd) {
 	    rd += s;
 	}
     }
-    c->offset = rd;
 
+    if (IS_CRAM_3_VERS(fd)) {
+	uint32_t crc, i;
+	unsigned char *dat = malloc(50 + 5*(c->num_landmarks)), *cp = dat;
+	if (!dat) {
+	    cram_free_container(c);
+	    return NULL;
+	}
+	if (-1 == int32_decode(fd, &c->crc32))return NULL; else rd+=4;
+
+	/* Reencode first as we can't easily access the original byte stream.
+	 *
+	 * FIXME: Technically this means this may not be fool proof. We could
+	 * create a CRAM file using a 2 byte ITF8 value that can fit in a 
+	 * 1 byte field, meaning the encoding is different to the original
+	 * form and so has a different CRC.
+	 *
+	 * The correct implementation would be to have an alternative form
+	 * of itf8_decode which also squirrels away the raw byte stream
+	 * during decoding so we can then CRC that.
+	 */
+	*(unsigned int *)cp = le_int4(c->length); cp += 4;
+	cp += itf8_put(cp, c->ref_seq_id);
+	cp += itf8_put(cp, c->ref_seq_start);
+	cp += itf8_put(cp, c->ref_seq_span);
+	cp += itf8_put(cp, c->num_records);
+	cp += itf8_put(cp, c->record_counter);
+	cp += itf8_put(cp, c->num_bases);
+	cp += itf8_put(cp, c->num_blocks);
+	cp += itf8_put(cp, c->num_landmarks);
+	for (i = 0; i < c->num_landmarks; i++) {
+	    cp += itf8_put(cp, c->landmark[i]);
+	}
+
+	crc = crc32(0L, dat, cp-dat);
+	if (crc != c->crc32) {
+	    fprintf(stderr, "Container header CRC32 failure\n");
+	    cram_free_container(c);
+	    return NULL;
+	}
+    }
+
+    c->offset = rd;
     c->slices = NULL;
     c->curr_slice = 0;
     c->max_slice = c->num_landmarks;
@@ -2194,11 +2557,11 @@ int cram_write_container(cram_fd *fd, cram_container *c) {
     char buf_a[1024], *buf = buf_a, *cp;
     int i;
 
-    if (50 + c->num_landmarks * 5 >= 1024)
-	buf = malloc(50 + c->num_landmarks * 5);
+    if (55 + c->num_landmarks * 5 >= 1024)
+	buf = malloc(55 + c->num_landmarks * 5);
     cp = buf;
 
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	cp += itf8_put(cp, c->length);
     } else {
 	*(int32_t *)cp = le_int4(c->length);
@@ -2214,7 +2577,7 @@ int cram_write_container(cram_fd *fd, cram_container *c) {
 	cp += itf8_put(cp, c->ref_seq_span);
     }
     cp += itf8_put(cp, c->num_records);
-    if (fd->version != CRAM_1_VERS) {
+    if (!IS_CRAM_1_VERS(fd)) {
 	cp += itf8_put(cp, c->record_counter);
 	cp += ltf8_put(cp, c->num_bases);
     }
@@ -2222,6 +2585,16 @@ int cram_write_container(cram_fd *fd, cram_container *c) {
     cp += itf8_put(cp, c->num_landmarks);
     for (i = 0; i < c->num_landmarks; i++)
 	cp += itf8_put(cp, c->landmark[i]);
+
+    if (IS_CRAM_3_VERS(fd)) {
+	c->crc32 = crc32(0L, buf, cp-buf);
+	cp[0] =  c->crc32        & 0xff;
+	cp[1] = (c->crc32 >>  8) & 0xff;
+	cp[2] = (c->crc32 >> 16) & 0xff;
+	cp[3] = (c->crc32 >> 24) & 0xff;
+	cp += 4;
+    }
+
     if (cp-buf != fwrite(buf, 1, cp-buf, fd->fp)) {
 	if (buf != buf_a)
 	    free(buf);
@@ -2711,11 +3084,9 @@ cram_file_def *cram_read_file_def(cram_fd *fd) {
 	return NULL;
     }
 
-    if (!(def->major_version == 1 && def->minor_version == 0) &&
-	!(def->major_version == 1 && def->minor_version == 1) &&
-	!(def->major_version == 2 && def->minor_version == 0)) {
+    if (def->major_version > 3) {
 	fprintf(stderr, "CRAM version number mismatch\n"
-		"Expected 1.0 or 2.0, got %d.%d\n",
+		"Expected 1.x, 2.x or 3.x, got %d.%d\n",
 		def->major_version, def->minor_version);
 	free(def);
 	return NULL;
@@ -2759,7 +3130,7 @@ SAM_hdr *cram_read_SAM_hdr(cram_fd *fd) {
     SAM_hdr *hdr;
 
     /* 1.1 onwards stores the header in the first block of a container */
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	/* Length */
 	if (-1 == int32_decode(fd, &header_len))
 	    return NULL;
@@ -2791,7 +3162,7 @@ SAM_hdr *cram_read_SAM_hdr(cram_fd *fd) {
 	    return NULL;
 	}
 
-	len = b->comp_size + 2 +
+	len = b->comp_size + 2 + 4*IS_CRAM_3_VERS(fd) +
 	    itf8_size(b->content_id) + 
 	    itf8_size(b->uncomp_size) + 
 	    itf8_size(b->comp_size);
@@ -2817,7 +3188,7 @@ SAM_hdr *cram_read_SAM_hdr(cram_fd *fd) {
 		cram_free_container(c);
 		return NULL;
 	    }
-	    len += b->comp_size + 2 + 
+	    len += b->comp_size + 2 + 4*IS_CRAM_3_VERS(fd) + 
 		itf8_size(b->content_id) + 
 		itf8_size(b->uncomp_size) + 
 		itf8_size(b->comp_size);
@@ -2890,7 +3261,7 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
     int header_len;
 
     /* 1.0 requires and UNKNOWN read-group */
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	if (!sam_hdr_find_rg(hdr, "UNKNOWN"))
 	    if (sam_hdr_add(hdr, "RG",
 			    "ID", "UNKNOWN", "SM", "UNKNOWN", NULL))
@@ -2915,8 +3286,11 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 		rlen = fd->refs->ref_id[i]->length;
 		MD5_Init(&md5);
 		ref = cram_get_ref(fd, i, 1, rlen);
+		if (NULL == ref) return -1;
+		rlen = fd->refs->ref_id[i]->length; /* In case it just loaded */
 		MD5_Update(&md5, ref, rlen);
 		MD5_Final(buf, &md5);
+		cram_ref_decr(fd->refs, i);
 
 		for (j = 0; j < 16; j++) {
 		    buf2[j*2+0] = "0123456789abcdef"[buf[j]>>4];
@@ -2925,7 +3299,6 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 		buf2[32] = 0;
 		if (sam_hdr_update(hdr, ty, "M5", buf2, NULL))
 		    return -1;
-		cram_ref_decr(fd->refs, i);
 	    }
 
 	    if (fd->ref_fn) {
@@ -2942,7 +3315,7 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 
     /* Length */
     header_len = sam_hdr_length(hdr);
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	if (-1 == int32_encode(fd, header_len))
 	    return -1;
 
@@ -2973,12 +3346,12 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 	    return -1;
 	c->landmark[0] = 0;
 
-	c->length = b->uncomp_size + 2 +
+	c->length = b->uncomp_size + 2 + 4*IS_CRAM_3_VERS(fd) +
 	    itf8_size(b->content_id)   + 
 	    itf8_size(b->uncomp_size)  +
 	    itf8_size(b->comp_size);
 #else
-	c->length = b->uncomp_size + 2 +
+	c->length = b->uncomp_size + 2 + 4*IS_CRAM_3_VERS(fd) +
 	    itf8_size(b->content_id)   + 
 	    itf8_size(b->uncomp_size)  +
 	    itf8_size(b->comp_size);
@@ -2997,7 +3370,7 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 #endif
 
 #ifdef PADDED_BLOCK
-	padded_length = MAX(c->length*2, 10000) - c->length;
+	padded_length = MAX(c->length*1.5, 10000) - c->length;
 	c->length += padded_length;
 	if (NULL == (pads = calloc(1, padded_length))) {
 	    cram_free_block(b);
@@ -3085,7 +3458,7 @@ static void cram_init_tables(cram_fd *fd) {
     fd->L2['T'] = 3; fd->L2['t'] = 3;
     fd->L2['N'] = 4; fd->L2['n'] = 4;
 
-    if (fd->version == CRAM_1_VERS) {
+    if (IS_CRAM_1_VERS(fd)) {
 	for (i = 0; i < 0x200; i++) {
 	    int f = 0;
 
@@ -3149,8 +3522,10 @@ static void cram_init_tables(cram_fd *fd) {
 }
 
 // Default version numbers for CRAM
+//static int major_version = 3;
+//static int minor_version = 0;
 static int major_version = 2;
-static int minor_version = 0;
+static int minor_version = 1;
 
 /*
  * Opens a CRAM file for read (mode "rb") or write ("wb").
@@ -3192,7 +3567,7 @@ cram_fd *cram_open(const char *filename, const char *mode) {
 	if (!(fd->file_def = cram_read_file_def(fd)))
 	    goto err;
 
-	fd->version = fd->file_def->major_version * 100 +
+	fd->version = fd->file_def->major_version * 256 +
 	    fd->file_def->minor_version;
 
 	if (!(fd->header = cram_read_SAM_hdr(fd)))
@@ -3213,7 +3588,7 @@ cram_fd *cram_open(const char *filename, const char *mode) {
 	if (0 != cram_write_file_def(fd, &def))
 	    goto err;
 
-	fd->version = def.major_version * 100 + def.minor_version;
+	fd->version = def.major_version * 256 + def.minor_version;
 
 	/* SAM header written later */
     }
@@ -3242,6 +3617,7 @@ cram_fd *cram_open(const char *filename, const char *mode) {
     fd->no_ref = 0;
     fd->ignore_md5 = 0;
     fd->use_bz2 = 0;
+    fd->use_arith = 0;
     fd->multi_seq = 0;
     fd->unsorted   = 0;
     fd->shared_ref = 0;
@@ -3253,7 +3629,7 @@ cram_fd *cram_open(const char *filename, const char *mode) {
     fd->job_pending = NULL;
     fd->ooc         = 0;
 
-    for (i = 0; i < 7; i++)
+    for (i = 0; i < 10; i++)
 	fd->m[i] = cram_new_metrics();
 
     fd->range.refid = -2; // no ref.
@@ -3307,7 +3683,6 @@ int cram_close(cram_fd *fd) {
     spare_bams *bl, *next;
     int i;
 	
-
     if (!fd)
 	return -1;
 
@@ -3335,6 +3710,32 @@ int cram_close(cram_fd *fd) {
 	t_results_queue_destroy(fd->rqueue);
     }
 
+    if (fd->mode == 'w') {
+	/* Write EOF block */
+	if (IS_CRAM_3_VERS(fd)) {
+	    if (1 != fwrite("\x0f\x00\x00\x00\xff\xff\xff\xff" // Cont HDR
+			    "\xff\xe0\x45\x4f\x46\x00\x00\x00" // Cont HDR
+			    "\x00\x01\x00"                     // Cont HDR
+			    "\x05\xbd\xd9\x4f"                 // CRC
+			    "\x00\x01\x00\x06\x06"             // Comp.HDR blk
+			    "\x01\x00\x01\x00\x01\x00"         // Comp.HDR blk
+			    "\xee\x63\x01\x4b",                // CRC
+			    38, 1, fd->fp))
+		return -1;
+	} else { 
+	    if (1 != fwrite("\x0b\x00\x00\x00\xff\xff\xff\xff"
+			    "\xff\xe0\x45\x4f\x46\x00\x00\x00"
+			    "\x00\x01\x00\x00\x01\x00\x06\x06"
+			    "\x01\x00\x01\x00\x01\x00", 30, 1, fd->fp))
+		return -1;
+	}		
+
+//	if (1 != fwrite("\x00\x00\x00\x00\xff\xff\xff\xff"
+//			"\xff\xe0\x45\x4f\x46\x00\x00\x00"
+//			"\x00\x00\x00", 19, 1, fd->fp))
+//	    return -1;
+    }
+
     for (bl = fd->bl; bl; bl = next) {
 	int i, max_rec = fd->seqs_per_slice * fd->slices_per_container;
 
@@ -3366,7 +3767,7 @@ int cram_close(cram_fd *fd) {
     if (fd->ref_free)
         free(fd->ref_free);
 
-    for (i = 0; i < 7; i++)
+    for (i = 0; i < 10; i++)
 	if (fd->m[i])
 	    free(fd->m[i]);
 
@@ -3456,6 +3857,10 @@ int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args) {
 	fd->use_bz2 = va_arg(args, int);
 	break;
 
+    case CRAM_OPT_USE_ARITH:
+	fd->use_arith = va_arg(args, int);
+	break;
+
     case CRAM_OPT_SHARED_REF:
 	fd->shared_ref = 1;
 	refs = va_arg(args, refs_t *);
@@ -3476,11 +3881,19 @@ int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args) {
 	return cram_load_reference(fd, va_arg(args, char *));
 
     case CRAM_OPT_VERSION: {
+	int major, minor;
 	char *s = va_arg(args, char *);
-	if (2 != sscanf(s, "%d.%d", &major_version, &minor_version)) {
+	if (2 != sscanf(s, "%d.%d", &major, &minor)) {
 	    fprintf(stderr, "Malformed version string %s\n", s);
 	    return -1;
 	}
+	if (!((major == 1 &&  minor == 0) ||
+	      (major == 2 && (minor == 0 || minor == 1)) ||
+	      (major == 3 &&  minor == 0))) {
+	    fprintf(stderr, "Unknown version string; "
+		    "use 1.0, 2.0, 2.1 or 3.0\n");
+	    return -1;
+	}
 	break;
     }
 
@@ -3520,6 +3933,10 @@ int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args) {
 	//t_pool_dispatch(fd->pool, cram_decoder_thread, fd);
 	break;
 
+    case CRAM_OPT_BINNING:
+	fd->binning = va_arg(args, int);
+	break;
+
     default:
 	fprintf(stderr, "Unknown CRAM option code %d\n", opt);
 	return -1;
