@@ -178,8 +178,8 @@ int cram_uncompress_block(cram_block *b);
  *        -1 on failure
  */
 int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
-			int level,  int strat,
-			int level2, int strat2);
+			enum cram_block_method m1, int level1, int strat1,
+			enum cram_block_method m2, int level2, int strat2);
 
 cram_metrics *cram_new_metrics(void);
 char *cram_block_method2str(enum cram_block_method m);
@@ -463,7 +463,9 @@ int cram_flush(cram_fd *fd);
 /*! Checks for end of file on a cram_fd stream.
  *
  * @return
- * Returns 1 if we hit an EOF while reading.
+ * Returns 0 if not at end of file
+ *         1 if we hit an expected EOF (end of range or EOF block)
+ *         2 for other EOF (end of stream without EOF block)
  */
 int cram_eof(cram_fd *fd);
 
