@@ -115,16 +115,23 @@ typedef struct {
     char    file_id[20];      // Filename or SHA1 checksum
 } cram_file_def;
 
-#define CRAM_1_VERS 100 // 1.0
-#define CRAM_2_VERS 200 // 1.1, or 2.0?
+#define CRAM_MAJOR_VERS(v) ((v) >> 8)
+#define CRAM_MINOR_VERS(v) ((v) & 0xff)
+#define IS_CRAM_1_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)==1)
+#define IS_CRAM_2_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)==2)
+#define IS_CRAM_3_VERS(fd) (CRAM_MAJOR_VERS((fd)->version)==3)
 
 struct cram_slice;
 
 enum cram_block_method {
-    RAW   = 0,
-    GZIP  = 1,
-    BZIP2 = 2,
-    ARITH1 = 3,
+    RAW    = 0,
+    GZIP   = 1,
+    BZIP2  = 2,
+    LZMA   = 3,
+    ARITH0 = 4,
+    ARITH1 = 5,
+    RANS0  = 6,
+    RANS1  = 7,
 };
 
 enum cram_content_type {
@@ -151,6 +158,7 @@ typedef struct {
     int32_t  content_id;
     int32_t  comp_size;
     int32_t  uncomp_size;
+    uint32_t crc32;
     int32_t  idx; /* offset into data */
     unsigned char    *data;
 
@@ -280,7 +288,7 @@ typedef struct {
 
     /* Size of container header above */
     size_t   offset;
-
+    
     /* Compression header is always the first block? */
     cram_block_compression_hdr *comp_hdr;
     cram_block *comp_hdr_block;
@@ -341,6 +349,8 @@ typedef struct {
     int *refs_used;       // array of frequency of ref seq IDs
 
     char *last_name;
+
+    uint32_t crc32;       // CRC32
 } cram_container;
 
 /*
@@ -547,6 +557,7 @@ typedef struct {
 
     pthread_mutex_t lock;  // Mutex for multi-threaded updating
     ref_entry *last;       // Last queried sequence
+    int last_id;           // Used in cram_ref_decr_locked to delay free
 } refs_t;
 
 /*-----------------------------------------------------------------------------
@@ -623,7 +634,7 @@ typedef struct {
 
     // compression level and metrics
     int level;
-    cram_metrics *m[7];
+    cram_metrics *m[10];
 
     // options
     int decode_md; // Whether to export MD and NM tags
@@ -634,7 +645,9 @@ typedef struct {
     int no_ref;
     int ignore_md5;
     int use_bz2;
+    int use_arith;
     int shared_ref;
+    enum quality_binning binning;
     cram_range range;
 
     // lookup tables, stored here so we can be trivially multi-threaded
@@ -651,6 +664,7 @@ typedef struct {
     int last_slice;                     // number of recs encoded in last slice
     int multi_seq;
     int unsorted;
+    int empty_container; 		// Marker for EOF block
     
     // thread pool
     int own_pool;
@@ -681,6 +695,8 @@ enum cram_option {
     CRAM_OPT_SHARED_REF,
     CRAM_OPT_NTHREADS,
     CRAM_OPT_THREAD_POOL,
+    CRAM_OPT_BINNING,
+    CRAM_OPT_USE_ARITH,
 };
 
 /* BF bitfields */
