@@ -1,3 +1,71 @@
+/*
+ * Copyright (c) 2003-2008, 2013 Genome Research Ltd.
+ * Author(s): James Bonfield, Andrew Whitwham
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *    2. Redistributions in binary form must reproduce the above
+ *       copyright notice, this list of conditions and the following
+ *       disclaimer in the documentation and/or other materials provided
+ *       with the distribution.
+ * 
+ *    3. Neither the names Genome Research Ltd and Wellcome Trust Sanger
+ *    Institute nor the names of its contributors may be used to endorse
+ *    or promote products derived from this software without specific
+ *    prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY GENOME RESEARCH LTD AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GENOME RESEARCH
+ * LTD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
+/*
+ * Author(s): James Bonfield
+ * 
+ * Copyright (c) 2001 MEDICAL RESEARCH COUNCIL
+ * All rights reserved
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1 Redistributions of source code must retain the above copyright notice, 
+ *      this list of conditions and the following disclaimer.
+ * 
+ *    2 Redistributions in binary form must reproduce the above copyright
+ *      notice, this list of conditions and the following disclaimer in
+ *      the documentation and/or other materials provided with the
+ *      distribution.
+ * 
+ *    3 Neither the name of the MEDICAL RESEARCH COUNCIL, THE LABORATORY OF
+ *      MOLECULAR BIOLOGY nor the names of its contributors may be used
+ *      to endorse or promote products derived from this software without
+ *      specific prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
+ */
+
 #ifdef HAVE_CONFIG_H
 #  include "io_lib_config.h"
 #endif
@@ -383,6 +451,7 @@ static mFILE *find_file_archive(char *file, char *arcname) {
 }
 #endif
 
+#ifndef SAMTOOLS
 #ifdef USE_WGET
 /* NB: non-reentrant due to reuse of handle */
 mFILE *find_file_url(char *file, char *url) {
@@ -420,17 +489,25 @@ mFILE *find_file_url(char *file, char *url) {
 }
 #endif
 
+#define STATIC_CURL
+
 #ifdef HAVE_LIBCURL
 mFILE *find_file_url(char *file, char *url) {
     char buf[8192], *cp;
     mFILE *mf = NULL, *headers = NULL;
     int maxlen = 8190 - strlen(file);
+#ifdef STATIC_CURL
     static CURL *handle = NULL;
     static int curl_init = 0;
+#else
+    CURL *handle = NULL;
+    static int curl_init = 0;
+#endif
     char errbuf[CURL_ERROR_SIZE];
 
     *errbuf = 0;
 
+#ifdef STATIC_CURL
     if (!curl_init) {
 	if (curl_global_init(CURL_GLOBAL_ALL))
 	    return NULL;
@@ -440,6 +517,17 @@ mFILE *find_file_url(char *file, char *url) {
 
 	curl_init = 1;
     }
+#else
+    if (!curl_init) {
+	if (curl_global_init(CURL_GLOBAL_ALL))
+	    return NULL;
+
+	curl_init = 1;
+    }
+    if (NULL == (handle = curl_easy_init()))
+	goto error;
+
+#endif
 
     /* Expand %s for the trace name */
     for (cp = buf; *url && cp - buf < maxlen; url++) {
@@ -459,7 +547,7 @@ mFILE *find_file_url(char *file, char *url) {
 
     if (0 != curl_easy_setopt(handle, CURLOPT_URL, buf))
 	goto error;
-    if (0 != curl_easy_setopt(handle, CURLOPT_TIMEOUT, 10L))
+    if (0 != curl_easy_setopt(handle, CURLOPT_CONNECTTIMEOUT, 60L))
 	goto error;
     if (0 != curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION,
 			      (curl_write_callback)mfwrite))
@@ -501,16 +589,24 @@ mFILE *find_file_url(char *file, char *url) {
 
     mfdestroy(headers);
 
+#ifndef STATIC_CURL
+    curl_easy_cleanup(handle);
+#endif
+
     mrewind(mf);
     return mf;
 
  error:
+#ifndef STATIC_CURL
+    if (handle)
+	curl_easy_cleanup(handle);
+#endif
     if (mf)
 	mfdestroy(mf);
     if (headers)
 	mfdestroy(headers);
     if (*errbuf)
-	fprintf(stderr, "%s\n", errbuf);
+	fprintf(stderr, "CURL ERROR: %s\n", errbuf);
     return NULL;
 }
 #endif
@@ -522,7 +618,6 @@ mFILE *find_file_url(char *file, char *url) {
 #endif
 
 
-#ifndef SAMTOOLS
 /*
  * Takes an SFF file in 'data' and edits the header to ensure
  * that it has no index listed and only claims to contain a single entry.
@@ -1118,6 +1213,7 @@ mFILE *open_path_mfile(char *file, char *path, char *relative_to) {
 		    return fp;
 		}
 #endif
+#ifndef SAMTOOLS
 #if defined(USE_WGET) || defined(HAVE_LIBCURL)
 	    } else if (0 == strncmp(ele2, "URL=", 4)) {
 		if (valid && (fp = find_file_url(file2, ele2+4))) {
@@ -1125,7 +1221,6 @@ mFILE *open_path_mfile(char *file, char *path, char *relative_to) {
 		    return fp;
 		}
 #endif
-#ifndef SAMTOOLS
 	    } else if (0 == strncmp(ele2, "SFF=", 4)) {
 		if (valid && (fp = find_file_sff(file2, ele2+4))) {
 		    free(newsearch);
