@@ -304,7 +304,9 @@ int scram_get_seq(scram_fd *fd, bam_seq_t **bsp) {
 	    return 0;
 
 	case 0:
-	    fd->eof = 1;
+	    // FIXME: if we ever implement range queries for BAM this will
+	    // need amendments to not claim a sub-range is invalid EOF.
+	    fd->eof = fd->b->eof_block ? 1 : 2;
 
 	default:
 	    return -1;
@@ -312,14 +314,12 @@ int scram_get_seq(scram_fd *fd, bam_seq_t **bsp) {
     }
 
     if (-1 == cram_get_bam_seq(fd->c, bsp)) {
-	if (cram_eof(fd->c))
-	    fd->eof = 1;
+	fd->eof = cram_eof(fd->c);
 	return -1;
     }
     return 0;
 }
 
-
 int scram_next_seq(scram_fd *fd, bam_seq_t **bsp) {
     return scram_get_seq(fd, bsp);
 }
@@ -356,6 +356,12 @@ int scram_set_option(scram_fd *fd, enum cram_option opt, ...) {
 	    fd->pool = NULL;
 	    return 0;
 	}
+    } else if (opt == CRAM_OPT_BINNING) {
+	int bin = va_arg(args, int);
+
+	return fd->is_bam
+	    ? bam_set_option (fd->b,  BAM_OPT_BINNING, bin)
+	    : cram_set_option(fd->c, CRAM_OPT_BINNING, bin);
     }
 
     if (!fd->is_bam) {
