@@ -45,6 +45,10 @@
 #include <ctype.h>
 #include <errno.h>
 
+#if defined(__MINGW32__)
+#   include <getopt.h>
+#endif
+
 #include <io_lib/scram.h>
 #include <io_lib/os.h>
 
@@ -105,6 +109,7 @@ static void usage(FILE *fp) {
     fprintf(fp, "    -j             [Cram] Compress using bzip2.\n");
 #endif
     fprintf(fp, "    -t N           Use N threads (availability varies by format)\n");
+    fprintf(fp, "    -B             Enable Illumina 8 quality-binning system (lossy)\n");
 }
 
 int main(int argc, char **argv) {
@@ -115,14 +120,15 @@ int main(int argc, char **argv) {
     int c, verbose = 0;
     int s_opt = 0, S_opt = 0, embed_ref = 0, ignore_md5 = 0, decode_md = 0;
     char *ref_fn = NULL;
-    int start, end, multi_seq = -1, no_ref = 0, use_bz2 = 0;
+    int start, end, multi_seq = -1, no_ref = 0, use_bz2 = 0, use_arith = 0;
     char ref_name[1024] = {0};
     refs_t *refs;
     int nthreads = 1;
     t_pool *p = NULL;
+    enum quality_binning binning = BINNING_NONE;
 
     /* Parse command line arguments */
-    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!Mmjt:")) != -1) {
+    while ((c = getopt(argc, argv, "u0123456789hvs:S:V:r:xXeI:O:R:!MmjJt:B")) != -1) {
 	switch (c) {
 	case '0': case '1': case '2': case '3': case '4':
 	case '5': case '6': case '7': case '8': case '9':
@@ -219,6 +225,10 @@ int main(int argc, char **argv) {
 #endif
 	    break;
 
+	case 'J':
+	    use_arith = 1;
+	    break;
+
 	case 't':
 	    nthreads = atoi(optarg);
 	    if (nthreads < 1) {
@@ -227,6 +237,10 @@ int main(int argc, char **argv) {
 	    }
 	    break;
 
+	case 'B':
+	    binning = BINNING_ILLUMINA;
+	    break;
+
 	case '?':
 	    fprintf(stderr, "Unrecognised option: -%c\n", optopt);
 	    usage(stderr);
@@ -301,6 +315,14 @@ int main(int argc, char **argv) {
 	if (scram_set_option(out, CRAM_OPT_USE_BZIP2, use_bz2))
 	    return 1;
 
+    if (use_arith)
+	if (scram_set_option(out, CRAM_OPT_USE_ARITH, use_arith))
+	    return 1;
+
+    if (binning != BINNING_NONE)
+	if (scram_set_option(out, CRAM_OPT_BINNING, binning))
+	    return 1;
+
     if (no_ref) {
 	if (scram_set_option(in, CRAM_OPT_NO_REF, no_ref))
 	    return 1;
@@ -396,11 +418,23 @@ int main(int argc, char **argv) {
     s = NULL;
 
     while (scram_get_seq(in, &s) >= 0) {
-	if (-1 == scram_put_seq(out, s))
+	if (-1 == scram_put_seq(out, s)) {
+	    fprintf(stderr, "Failed to encode sequence\n");
 	    return 1;
+	}
     }
-    if (!scram_eof(in))
+    switch(scram_eof(in)) {
+    case 0:
+	fprintf(stderr, "Failed to decode sequence\n");
 	return 1;
+    case 2:
+	fprintf(stderr, "Warning: no end-of-file block identified. "
+		"File may be truncated.\n");
+	break;
+    case 1: default:
+	// expected case
+	break;
+    }
 
     /* Finally tidy up and close files */
     if (scram_close(in))
